// Libraries
import { Injectable } from '@angular/core';
import { ROOT_EFFECTS_INIT, Actions, Effect, ofType } from '@ngrx/effects';
import { Observable, forkJoin, of } from 'rxjs';
import { map, flatMap, tap } from 'rxjs/operators';

// Services
import { DegreePlannerApiService } from '@app/degree-planner/services/api.service';

// Actions
import {
	InitialPlanLoadResponse,
	ChangeVisiblePlanRequest,
	ChangeVisiblePlanResponse,
	PlanActionTypes
} from '@app/degree-planner/actions/plan.actions';

// Models
import { PlannedTerm } from '@app/core/models/planned-term';

@Injectable()
export class DegreePlanEffects {
	constructor(
		private actions$: Actions,
		private api: DegreePlannerApiService
	) {}

	@Effect()
	init$: Observable<InitialPlanLoadResponse> = this.actions$.pipe(
		ofType(ROOT_EFFECTS_INIT),

		// Load all plans available to the user.
		flatMap(() => this.api.getAllDegreePlans()),

		// Pick one of the plans to use as the visible plan.
		map(allDegreePlans => {
			const visibleRoadmapId = (
				allDegreePlans.find(plan => plan.primary) || allDegreePlans[0]
			).roadmapId;
			return { visibleRoadmapId, allDegreePlans };
		}),

		// Get term data for the degree plan specified by the roadmap ID and any
		// courses that were 'saved for later' by the user.
		flatMap(stdin => {
			return forkJoin(
				this.loadTermsForPlan(stdin),
				this.api.getSavedForLaterCourses()
			).pipe(
				map(([planDetails, savedForLaterCourses]) => {
					return { ...planDetails, savedForLaterCourses };
				})
			);
		}),

		// Wrap data in an Action for dispatch
		map(stdin => new InitialPlanLoadResponse(stdin))
	);

	@Effect()
	switch$: Observable<ChangeVisiblePlanResponse> = this.actions$.pipe(
		ofType<ChangeVisiblePlanRequest>(PlanActionTypes.ChangeVisiblePlanRequest),

		// Get the roadmap ID from the action.
		map(action => ({ visibleRoadmapId: action.payload.newVisibleRoadmapId })),

		// Get term data for the degree plan specified by the roadmap ID.
		flatMap(stdin => this.loadTermsForPlan(stdin)),

		// Wrap data in an Action for dispatch
		map(stdin => new ChangeVisiblePlanResponse(stdin))
	);

	private loadTermsForPlan<T extends { visibleRoadmapId: number }>(stdin: T) {
		return forkJoin(
			this.api.getAllNotes(stdin.visibleRoadmapId),
			this.api.getAllCourses(stdin.visibleRoadmapId),
			this.api.getActiveTerms()
		).pipe(
			// Combine courses and notes by term.
			map(([notes, courses, activeTerms]) => {
				/**
				 * Using the notes & courses relevant to the current degree plan and
				 * the active terms, generate a sorted list of all unqiue term codes.
				 */
				const uniqueTermCodes = [notes, courses, activeTerms]
					.map((ts: { termCode: string }[]) => ts.map(t => t.termCode))
					.reduce((flat, nested) => flat.concat(nested), [])
					.filter((termCode, index, self) => self.indexOf(termCode) === index)
					.sort();

				/**
				 * For each unique termCode, build a Term object that includes any
				 * courses or notes relevant to that termCode.
				 */
				const visibleTerms: PlannedTerm[] = uniqueTermCodes.map(termCode => {
					return {
						termCode,
						note: notes.find(note => note.termCode === termCode),
						courses: courses.filter(course => course.termCode === termCode)
					};
				});

				return Object.assign({}, stdin, { visibleTerms, activeTerms });
			})
		);
	}
}
